"""
Wrappers for all workflow tasks
"""
import json
from abc import ABC
from abc import abstractmethod
from pathlib import Path
from typing import Generator
from typing import List
from typing import Optional
from typing import Union

from astropy.io import fits
from dkist_processing_core import TaskBase

from dkist_processing_common._util.graphql import add_new_recipe_run_status
from dkist_processing_common._util.graphql import apply_status_id_to_recipe_run
from dkist_processing_common._util.graphql import get_message_status_query
from dkist_processing_common._util.graphql import graph_ql_client
from dkist_processing_common._util.graphql import RecipeRunInputDatasetQuery
from dkist_processing_common._util.graphql import RecipeRunResponse
from dkist_processing_common._util.interservice_bus import CatalogFrameMessage
from dkist_processing_common._util.interservice_bus import CatalogObjectMessage
from dkist_processing_common._util.scratch import WorkflowFileSystem


class TaskBaseExt(TaskBase, ABC):
    """
    Wrapper for all tasks. Contains the data access layer.
    """

    def __init__(
        self,
        recipe_run_id: int,
        workflow_name: str,
        workflow_version: str,
        is_task_manual: bool = False,
        apm_client_config: Optional[dict] = None,
        scratch_base_path: Optional[Path] = None,
    ):
        self.scratch_base_path = scratch_base_path
        super().__init__(
            recipe_run_id, workflow_name, workflow_version, is_task_manual, apm_client_config
        )

    def set_scratch_base_path(self, path):
        self.scratch_base_path = path

    @property
    def _scratch(self):  # , scratch_base_path: Union[Path, str, None] = None):
        """
        WorkflowFileSystem object that provides the low level data access API
        """
        return WorkflowFileSystem(scratch_base_path=self.scratch_base_path)

    @property
    def input_dataset_path(self):
        """
        Path to the input dataset in scratch
        """
        return self._scratch.workflow_base_path / "input_dataset.json"

    @property
    def output_paths(self):
        """
        Paths of all objects tagged as "OUTPUT"
        """
        return self._scratch.find_all(tags=["OUTPUT"])

    def input_dataset(self, section: str = "all") -> Union[dict, None]:
        """
        Input dataset document for the recipe run
        """
        if not self.input_dataset_path.exists():
            # Get input dataset from db
            input_dataset_response = graph_ql_client.execute_gql_query(
                query_base="recipeRuns",
                query_response_cls=RecipeRunResponse,
                query_parameters=RecipeRunInputDatasetQuery(recipeRunId=self.recipe_run_id),
            )
            # Write document to disk for future use
            self._write_input_dataset(
                input_dataset_response[0].recipeInstance.inputDataset.inputDatasetDocument
            )

        input_dataset_document = self._read_input_dataset()

        section = section.lower()
        if section == "all":
            return input_dataset_document
        return input_dataset_document.get(section, None)

    def _write_input_dataset(self, input_dataset):
        """
        Interface for writing the input dataset document to scratch
        """
        self._scratch.write(
            bytes(input_dataset, "utf-8"), self.input_dataset_path.parts[-1], tags=["INPUTDATASET"]
        )

    def _read_input_dataset(self) -> dict:
        """
        Interface for reading the input dataset document from scratch
        """
        with self.input_dataset_path.open("r") as f:
            return json.loads(f.read())

    def read_input_fits(self, datatype: str) -> Generator[fits.HDUList, None, None]:
        """
        Get paths for all data tagged with "INPUT" and the datatype requested
        """
        return self.read_fits(tags=["INPUT", datatype.upper()])

    def write_intermediate_fits(
        self, data: fits.HDUList, datatype: str, tags: Optional[List] = None
    ) -> None:
        """
        Write a calibration product tagged with "INTERMEDIATE", the datatype supplied,
        and any extra tags supplied
        """
        relative_path = f"{datatype}.fits"
        full_tags = ["INTERMEDIATE", datatype.upper()]
        if tags:
            for tag in tags:
                full_tags.append(tag)
        self._scratch.write_fits(data=data, relative_path=relative_path, tags=full_tags)

    def write_output_fits(
        self, data: fits.HDUList, relative_path: Union[str, Path], tags: Optional[List] = None
    ) -> None:
        """
        Write an output frame tagged with "OUTPUT", "OBSERVE", "L1", and any extra tags supplied
        """
        # TODO how are filenames determined for output data?
        full_tags = ["OUTPUT", "OBSERVE", "L1"]
        if tags:
            for tag in tags:
                full_tags.append(tag)
        self._scratch.write_fits(data=data, relative_path=relative_path, tags=full_tags)

    def read_intermediate_fits(self, datatype: str):
        """
        Get paths for all data tagged with "INTERMEDIATE" and the datatype requested
        """
        try:
            paths = self._scratch.find_all(["INTERMEDIATE", datatype.upper()])
            filepath = next(paths)
        except StopIteration:
            raise FileNotFoundError(f"No files found of type {datatype}")
        try:
            next(paths)
        except StopIteration:
            return fits.open(filepath)
        raise RuntimeError(f"More than one file found of type {datatype}")

    def read_fits(self, tags: Union[List, str]):
        filepaths = self._scratch.find_all(tags=tags)
        return (fits.open(filepath) for filepath in filepaths)

    def write_movie(
        self, data
    ):  # TODO Write this method and potentially move it to the movie writing class in 'common'
        pass


class SupportTaskBase(TaskBaseExt, ABC):
    """
    Wrapper for all support tasks (tasks that do not impact the data)
    """

    recipe_run_statuses = {
        "INPROGRESS": "Recipe run is currently undergoing processing",
        "COMPLETEDSUCCESSFULLY": "Recipe run processing completed with no errors",
    }

    def change_status_to_in_progress(self):
        """
        Set the recipe run status to "INPROGRESS"
        """
        self._change_status(status="INPROGRESS", is_complete=False)

    def change_status_to_completed_successfully(self):
        """
        Set the recipe run status to "COMPLETEDSUCCESSFULLY"
        """
        self._change_status(status="COMPLETEDSUCCESSFULLY", is_complete=True)

    def _change_status(self, status: str, is_complete: bool):
        """
        Change the recipe run status of a recipe run to the given status
        """
        status_response = get_message_status_query(status=status)

        # If the status was found
        if len(status_response) > 0:
            # Get the status ID
            recipe_run_status_id = status_response[0].recipeRunStatusId
        else:
            # Add the status to the db and get the new status ID
            recipe_run_status_id = add_new_recipe_run_status(status=status, is_complete=is_complete)

        apply_status_id_to_recipe_run(
            recipe_run_status_id=recipe_run_status_id, recipe_run_id=self.recipe_run_id
        )

    @property
    def proposal_id(self) -> str:
        """
        Get the propoal id of the current recipe run
        """
        try:
            getattr(self, "_proposal_id")
        except AttributeError:
            self._proposal_id = graph_ql_client.execute_gql_query(
                query_base="recipeRuns",
                query_response_cls=RecipeRunResponse,
                query_parameters=RecipeRunInputDatasetQuery(recipeRunId=self.recipe_run_id),
            )[0].recipeInstance.processingCandidate.proposalId
        return self._proposal_id

    def create_frame_message(self, object_filepath: str):
        catalog_frame_message = CatalogFrameMessage(
            objectName=object_filepath, conversationId=str(self.recipe_run_id)
        )
        return catalog_frame_message

    def create_movie_message(self, object_filepath: str):
        catalog_movie_message = CatalogObjectMessage(
            objectType="MOVIE",
            objectName=object_filepath,
            groupId=self.dataset_id,
            conversationId=str(self.recipe_run_id),
        )
        return catalog_movie_message


class ScienceTaskL0ToL1(TaskBaseExt):
    """
    Wrapper for all science tasks creating L1 data from L0 raw inputs
    """

    def record_provenance(self):
        pass  # TODO

    @property
    def input_dark_frames(self) -> Generator[fits.HDUList, None, None]:
        """
        Get the list of input dark frames
        """
        return self.read_input_fits(datatype="DARK")

    @property
    def input_gain_frames(self) -> Generator[fits.HDUList, None, None]:
        """
        Get the list of input gain frames
        """
        return self.read_input_fits(datatype="GAIN")

    @property
    def input_target_frames(self) -> Generator[fits.HDUList, None, None]:
        """
        Get the list of input target frames
        """
        return self.read_input_fits(datatype="TARGET")

    @property
    def input_instpolcal_frames(self) -> Generator[fits.HDUList, None, None]:
        """
        Get the list of input instpolcal frames
        """
        return self.read_input_fits(datatype="INSTPOLCAL")

    @property
    def input_telpolcal_frames(self) -> Generator[fits.HDUList, None, None]:
        """
        Get the list of input telpolcal frames
        """
        return self.read_input_fits(datatype="TELPOLCAL")

    @property
    def input_geometric_frames(self) -> Generator[fits.HDUList, None, None]:
        """
        Get the list of input geometric frames
        """
        return self.read_input_fits(datatype="GEOMETRIC")

    def write_dark(self, data: fits.HDUList, tags: Optional[List] = None) -> None:
        """
        Write the calibrated dark frame
        """
        self.write_intermediate_fits(data=data, datatype="DARK", tags=tags)

    def write_gain(self, data: fits.HDUList, tags: Optional[List] = None) -> None:
        """
        Write the calibrated gain frame
        """
        self.write_intermediate_fits(data=data, datatype="GAIN", tags=tags)

    def write_target(self, data: fits.HDUList, tags: Optional[List] = None) -> None:
        """
        Write the calibrated target frame
        """
        self.write_intermediate_fits(data=data, datatype="TARGET", tags=tags)

    def write_instpolcal(self, data: fits.HDUList, tags: Optional[List] = None) -> None:
        """
        Write the calibrated instpolcal frame
        """
        self.write_intermediate_fits(data=data, datatype="INSTPOLCAL", tags=tags)

    def write_telpolcal(self, data: fits.HDUList, tags: Optional[List] = None) -> None:
        """
        Write the calibrated telpolcal frame
        """
        self.write_intermediate_fits(data=data, datatype="TELPOLCAL", tags=tags)

    def write_geometric(self, data: fits.HDUList, tags: Optional[List] = None) -> None:
        """
        Write the calibrated geometric frame
        """
        self.write_intermediate_fits(data=data, datatype="GEOMETRIC", tags=tags)

    @property
    def intermediate_dark(self) -> fits.HDUList:
        """
        Get the calibrated dark frame
        """
        return self.read_intermediate_fits(datatype="DARK")

    @property
    def intermediate_gain(self) -> fits.HDUList:
        """
        Get the calibrated gain frame
        """
        return self.read_intermediate_fits(datatype="GAIN")

    @property
    def intermediate_target(self) -> fits.HDUList:
        """
        Get the calibrated target frame
        """
        return self.read_intermediate_fits(datatype="TARGET")

    @property
    def intermediate_instpolcal(self) -> fits.HDUList:
        """
        Get the calibrated instpolcal frame
        """
        return self.read_intermediate_fits(datatype="INSTPOLCAL")

    @property
    def intermediate_telpolcal(self) -> fits.HDUList:
        """
        Get the calibrated telpolcal frame
        """
        return self.read_intermediate_fits(datatype="TELPOLCAL")

    @property
    def intermediate_geometric(self) -> fits.HDUList:
        """
        Get the calibrated geometric frame
        """
        return self.read_intermediate_fits(datatype="GEOMETRIC")

    @abstractmethod
    def run(self) -> None:
        """
        Abstract method that must be overridden to execute the desired DAG task.
        """



from pathlib import Path

import pytest

from dkist_processing_common._util.globus import submit_globus_transfer


def test_submit_globus_transfer_non_equal_source_and_dest():
    with pytest.raises(ValueError):
        submit_globus_transfer(
            source_files=[Path("1"), Path("2"), Path("3")],
            destination_files=[Path("4"), Path("5")],
            source_endpoint="test",
            destination_endpoint="test",
        )


def test_invalid_source_endpoint():
    with pytest.raises(ValueError):
        submit_globus_transfer(
            source_files=[Path("1"), Path("2"), Path("3")],
            destination_files=[Path("4"), Path("5"), Path("6")],
            source_endpoint="test",
            destination_endpoint="a62e451d-1927-4e74-9a8c-9c42d0bc5217",
        )


def test_invalid_destination_endpoint():
    with pytest.raises(ValueError):
        submit_globus_transfer(
            source_files=[Path("1"), Path("2"), Path("3")],
            destination_files=[Path("4"), Path("5"), Path("6")],
            source_endpoint="a62e451d-1927-4e74-9a8c-9c42d0bc5217",
            destination_endpoint="test",
        )


def test_globus_transfer(mocker):
    mocker.patch("dkist_processing_common._util.globus.get_globus_transfer_client", autospec=True)
    submit_globus_transfer(
        source_files=[Path("1"), Path("2"), Path("3")],
        destination_files=[Path("4"), Path("5"), Path("6")],
        source_endpoint="a62e451d-1927-4e74-9a8c-9c42d0bc5217",
        destination_endpoint="a62e451d-1927-4e74-9a8c-9c42d0bc5217",
    )



import pytest

from dkist_processing_common._util.interservice_bus import CatalogFrameMessage
from dkist_processing_common._util.interservice_bus import CatalogObjectMessage
from dkist_processing_common._util.message import publish_messages


def test_publish_messages(mocker):
    mocker.patch("dkist_processing_common._util.message.DurableBlockingProducerWrapper")
    messages = [CatalogFrameMessage(), CatalogObjectMessage(objectType="MOVIE")]
    publish_messages(messages)


def test_invalid_publish_messages(mocker):
    mocker.patch("dkist_processing_common._util.message.DurableBlockingProducerWrapper")
    messages = ["a"]
    with pytest.raises(AttributeError):
        publish_messages(messages)



from typing import List
from typing import Union

from talus import DurableBlockingProducerWrapper

from dkist_processing_common._util.interservice_bus import CatalogFrameMessage
from dkist_processing_common._util.interservice_bus import CatalogObjectMessage
from dkist_processing_common._util.interservice_bus import RABBITMQ_CONFIG


def publish_messages(messages: List[Union[CatalogFrameMessage, CatalogObjectMessage]]):
    """
    Send all frame, movie, and object messages for downstream services
    :param messages: the messages to be sent
    """
    bindings = [message.binding() for message in messages]
    with DurableBlockingProducerWrapper(
        producer_queue_bindings=bindings,
        publish_exchange="master.direct.x",
        retry_tries=0,
        **RABBITMQ_CONFIG
    ) as producer:

        for message in messages:
            producer.publish_message(message)



import pytest

from dkist_processing_common._util.graphql import CreateRecipeRunStatusResponse
from dkist_processing_common._util.graphql import RecipeRunStatusResponse


class FakeGQLClient:
    @staticmethod
    def execute_gql_mutation(**kwargs):
        mutation_base = kwargs["mutation_base"]

        if mutation_base == "updateRecipeRun":
            return
        if mutation_base == "createRecipeRunStatus":
            return CreateRecipeRunStatusResponse(
                recipeRunStatus=RecipeRunStatusResponse(recipeRunStatusId=1)
            )


def test_add_new_recipe_run_status(mocker):
    from dkist_processing_common._util.graphql import add_new_recipe_run_status

    mocker.patch("dkist_processing_common._util.graphql.graph_ql_client", new=FakeGQLClient)
    with pytest.raises(TypeError):
        add_new_recipe_run_status(status=4, is_complete=True)
    with pytest.raises(TypeError):
        add_new_recipe_run_status(status="INPROGRESS", is_complete="true")
    with pytest.raises(KeyError):
        add_new_recipe_run_status(status="READY", is_complete=True)
    assert isinstance(add_new_recipe_run_status(status="INPROGRESS", is_complete=True), int)



from pathlib import Path

import pytest
from astropy.io import fits

from dkist_processing_common._util.graphql import CreateRecipeRunStatusResponse
from dkist_processing_common._util.graphql import InputDatasetResponse
from dkist_processing_common._util.graphql import ProcessingCandidateResponse
from dkist_processing_common._util.graphql import RecipeInstanceResponse
from dkist_processing_common._util.graphql import RecipeRunResponse
from dkist_processing_common._util.graphql import RecipeRunStatusResponse


class FakeGQLClient:
    @staticmethod
    def execute_gql_query(**kwargs):
        query_base = kwargs["query_base"]

        if query_base == "recipeRunStatuses":
            return [RecipeRunStatusResponse(recipeRunStatusId=1)]
        if query_base == "recipeRuns":
            return [
                RecipeRunResponse(
                    recipeInstance=RecipeInstanceResponse(
                        processingCandidate=ProcessingCandidateResponse(
                            observingProgramExecutionId="abc", proposalId="123"
                        ),
                        inputDataset=InputDatasetResponse(
                            inputDatasetDocument='{"bucket": "bucket-name", "parameters": [{"parameterName": "", "parameterValues": [{"parameterValueId": 1, "parameterValue": "[[1,2,3],[4,5,6],[7,8,9]]", "parameterValueStartDate": "1/1/2000"}]}], "frames": ["objectKey1", "objectKey2", "objectKeyN"]}'
                        ),
                    )
                )
            ]

    @staticmethod
    def execute_gql_mutation(**kwargs):
        mutation_base = kwargs["mutation_base"]

        if mutation_base == "updateRecipeRun":
            return
        if mutation_base == "createRecipeRunStatus":
            return CreateRecipeRunStatusResponse(
                recipeRunStatus=RecipeRunStatusResponse(recipeRunStatusId=1)
            )


def test_change_status_to_in_progress(support_task, mocker):
    """
    Given: a support task
    When: requesting that the status of its recipe run is changed
    Then: the recipe run status changes
    """
    mocker.patch("dkist_processing_common._util.graphql.graph_ql_client", new=FakeGQLClient)
    support_task.change_status_to_in_progress()


def test_change_status_to_completed_successfully(support_task, mocker):
    """
    Given: a support task
    When: requesting that the status of its recipe run is changed
    Then: the recipe run status changes
    """
    mocker.patch("dkist_processing_common._util.graphql.graph_ql_client", new=FakeGQLClient)
    support_task.change_status_to_completed_successfully()


def test_input_dataset(support_task, mocker):
    """
    Given: a support task
    When: requesting an input data set
    Then: the input data set is returned either from a db query or local storage
    """
    mocker.patch("dkist_processing_common.base.graph_ql_client", new=FakeGQLClient)
    r_all = support_task.input_dataset()
    r_section = support_task.input_dataset(section="parameters")
    assert r_all["bucket"] == "bucket-name"
    assert len(r_all["frames"]) == 3
    assert r_section[0]["parameterValues"][0]["parameterValueId"] == 1
    assert r_section[0]["parameterValues"][0]["parameterValue"] == "[[1,2,3],[4,5,6],[7,8,9]]"
    assert r_section[0]["parameterValues"][0]["parameterValueStartDate"] == "1/1/2000"


def test_get_proposal_id(support_task, mocker):
    """
    Given: a support task
    When: the proposal id is requested
    Then: the proposal task is obtained via a db query and returned
    """
    mocker.patch("dkist_processing_common.base.graph_ql_client", new=FakeGQLClient)
    assert support_task.proposal_id == "123"


def test_frame_message(support_task):
    """
    Given: a support task
    When: a frame message is created
    Then: the components of the frame message are correctly populated
    """
    msg = support_task.create_frame_message(object_filepath="/test/object/path.ext")
    assert msg.objectName == "/test/object/path.ext"
    assert msg.conversationId == str(support_task.recipe_run_id)
    assert msg.bucket == "data"
    assert msg.incrementDatasetCatalogReceiptCount


def test_movie_message(support_task):
    """
    Given: a support task
    When: a movie message is created
    Then: the components of the movie message are correctly populated
    """
    msg = support_task.create_movie_message(object_filepath="/test/object/path.ext")
    assert msg.objectName == "/test/object/path.ext"
    assert msg.conversationId == str(support_task.recipe_run_id)
    assert msg.bucket == "data"
    assert msg.objectType == "MOVIE"
    assert msg.groupName == "DATASET"
    assert msg.incrementDatasetCatalogReceiptCount


def test_write_intermediate_fits(science_task, intermediate_fits, tmp_path):
    """
    Given: a fits file
    When: writing the file
    Then: the file exists
    """
    science_task.set_scratch_base_path(tmp_path)
    science_task.write_intermediate_fits(
        data=intermediate_fits, datatype="TEST_WRITE", tags=["INT1", "INT2"]
    )
    assert Path(science_task._scratch.workflow_base_path, "TEST_WRITE.fits").exists()


def test_write_output_fits(science_task, output_fits, tmp_path):
    """
    Given: a fits file
    When: writing the file
    Then: the file exists
    """
    relative_path = "test_output.fits"
    science_task.set_scratch_base_path(tmp_path)
    science_task.write_output_fits(
        data=output_fits, relative_path=relative_path, tags=["OUTPUT1", "OUTPUT2"]
    )
    assert Path(science_task._scratch.workflow_base_path, relative_path).exists()


def test_read_intermediate_fits(science_task, intermediate_fits, tmp_path):
    """
    Given: a type of file
    When: reading that file
    Then: the file is read correctly and the contents can be examined
    """
    science_task.set_scratch_base_path(tmp_path)
    science_task.write_intermediate_fits(data=intermediate_fits, datatype="TEST_READ")
    hdul = science_task.read_intermediate_fits(datatype="TEST_READ")
    assert hdul[0].header["DKIST004"] == "TEST_INTERMEDIATE"


@pytest.mark.parametrize(
    "tags",
    [
        pytest.param(["INPUT", "DARK"], id="DARK"),
        pytest.param(["INPUT", "GAIN"], id="GAIN"),
        pytest.param(["INPUT", "TARGET"], id="TARGET"),
        pytest.param(["INPUT", "INSTPOLCAL"], id="INSTPOLCAL"),
        pytest.param(["INPUT", "TELPOLCAL"], id="TELPOLCAL"),
        pytest.param(["INPUT", "GEOMETRIC"], id="GEOMETRIC"),
    ],
)
def test_read_fits(science_task, populate_input_fits, tags):
    """
    Given: a type of file
    When: reading that file
    Then: the file is read correctly and the contents can be examined
    """
    science_task.set_scratch_base_path(populate_input_fits)
    frames = science_task.read_fits(tags=tags)
    frame = next(frames)
    assert frame[0].header["DKIST004"] == tags[1]


def test_input_dark_frames(science_task, populate_input_fits):
    """
    Given: a specific type of input
    When: reading frames that match the input
    Then: the correct number of frames are read and the contents can be examined
    """
    science_task.set_scratch_base_path(populate_input_fits)
    frames = science_task.input_dark_frames
    frame = next(frames)
    assert frame[0].header["DKIST004"] == "DARK"


def test_input_gain_frames(science_task, populate_input_fits):
    """
    Given: a specific type of input
    When: reading frames that match the input
    Then: the correct number of frames are read and the contents can be examined
    """
    science_task.set_scratch_base_path(populate_input_fits)
    frames = science_task.input_gain_frames
    frame = next(frames)
    assert frame[0].header["DKIST004"] == "GAIN"


def test_input_target_frames(science_task, populate_input_fits):
    """
    Given: a specific type of input
    When: reading frames that match the input
    Then: the correct number of frames are read and the contents can be examined
    """
    science_task.set_scratch_base_path(populate_input_fits)
    frames = science_task.input_target_frames
    frame = next(frames)
    assert frame[0].header["DKIST004"] == "TARGET"


def test_input_instpolcal_frames(science_task, populate_input_fits):
    """
    Given: a specific type of input
    When: reading frames that match the input
    Then: the correct number of frames are read and the contents can be examined
    """
    science_task.set_scratch_base_path(populate_input_fits)
    frames = science_task.input_instpolcal_frames
    frame = next(frames)
    assert frame[0].header["DKIST004"] == "INSTPOLCAL"


def test_input_telpolcal_frames(science_task, populate_input_fits):
    """
    Given: a specific type of input
    When: reading frames that match the input
    Then: the correct number of frames are read and the contents can be examined
    """
    science_task.set_scratch_base_path(populate_input_fits)
    frames = science_task.input_telpolcal_frames
    frame = next(frames)
    assert frame[0].header["DKIST004"] == "TELPOLCAL"


def test_input_geometric_frames(science_task, populate_input_fits):
    """
    Given: a specific type of input
    When: reading frames that match the input
    Then: the correct number of frames are read and the contents can be examined
    """
    science_task.set_scratch_base_path(populate_input_fits)
    frames = science_task.input_geometric_frames
    frame = next(frames)
    assert frame[0].header["DKIST004"] == "GEOMETRIC"


def test_intermediate_dark(science_task, tmp_path):
    """
    Given: a specific type of input
    When: reading frames that match the input
    Then: the correct number of frames are read and the contents can be examined
    """
    datatype = "DARK"
    hdu = fits.PrimaryHDU()
    hdu.header["datatype"] = datatype
    hdul = fits.HDUList([hdu])
    science_task.set_scratch_base_path(tmp_path)
    science_task.write_dark(data=hdul)
    assert science_task.intermediate_dark[0].header["datatype"] == datatype


def test_intermediate_gain(science_task, tmp_path):
    """
    Given: a specific type of input
    When: reading frames that match the input
    Then: the correct number of frames are read and the contents can be examined
    """
    datatype = "GAIN"
    hdu = fits.PrimaryHDU()
    hdu.header["datatype"] = datatype
    hdul = fits.HDUList([hdu])
    science_task.set_scratch_base_path(tmp_path)
    science_task.write_gain(data=hdul)
    assert science_task.intermediate_gain[0].header["datatype"] == datatype


def test_intermediate_target(science_task, tmp_path):
    """
    Given: a specific type of input
    When: reading frames that match the input
    Then: the correct number of frames are read and the contents can be examined
    """
    datatype = "TARGET"
    hdu = fits.PrimaryHDU()
    hdu.header["datatype"] = datatype
    hdul = fits.HDUList([hdu])
    science_task.set_scratch_base_path(tmp_path)
    science_task.write_target(data=hdul)
    assert science_task.intermediate_target[0].header["datatype"] == datatype


def test_intermediate_instpolcal(science_task, tmp_path):
    """
    Given: a specific type of input
    When: reading frames that match the input
    Then: the correct number of frames are read and the contents can be examined
    """
    datatype = "INSTPOLCAL"
    hdu = fits.PrimaryHDU()
    hdu.header["datatype"] = datatype
    hdul = fits.HDUList([hdu])
    science_task.set_scratch_base_path(tmp_path)
    science_task.write_instpolcal(data=hdul)
    assert science_task.intermediate_instpolcal[0].header["datatype"] == datatype


def test_intermediate_telpolcal(science_task, tmp_path):
    """
    Given: a specific type of input
    When: reading frames that match the input
    Then: the correct number of frames are read and the contents can be examined
    """
    datatype = "TELPOLCAL"
    hdu = fits.PrimaryHDU()
    hdu.header["datatype"] = datatype
    hdul = fits.HDUList([hdu])
    science_task.set_scratch_base_path(tmp_path)
    science_task.write_telpolcal(data=hdul)
    assert science_task.intermediate_telpolcal[0].header["datatype"] == datatype


def test_intermediate_geometric(science_task, tmp_path):
    """
    Given: a specific type of input
    When: reading frames that match the input
    Then: the correct number of frames are read and the contents can be examined
    """
    datatype = "GEOMETRIC"
    hdu = fits.PrimaryHDU()
    hdu.header["datatype"] = datatype
    hdul = fits.HDUList([hdu])
    science_task.set_scratch_base_path(tmp_path)
    science_task.write_geometric(data=hdul)
    assert science_task.intermediate_geometric[0].header["datatype"] == datatype



import uuid
from os import environ
from pathlib import Path
from typing import List

from globus_sdk import ClientCredentialsAuthorizer
from globus_sdk import ConfidentialAppAuthClient
from globus_sdk import GlobusError
from globus_sdk import TransferClient
from globus_sdk import TransferData


def get_globus_transfer_client() -> TransferClient:
    confidential_client = ConfidentialAppAuthClient(
        client_id=environ.get("GLOBUS_CLIENT_ID"), client_secret=environ.get("GLOBUS_CLIENT_SECRET")
    )
    authorizer = ClientCredentialsAuthorizer(
        confidential_client, scopes="urn:globus:auth:scope:transfer.api.globus.org:all"
    )
    return TransferClient(authorizer=authorizer)


def create_transfer(tc, source_endpoint, destination_endpoint) -> TransferData:
    return TransferData(
        tc,
        source_endpoint,
        destination_endpoint,
        verify_checksum=True,
    )


def submit_globus_transfer(
    source_files: List[Path],
    destination_files: List[Path],
    source_endpoint: str,
    destination_endpoint: str,
):
    """
    Parse the lists of source and destination filepaths, and submit the transfer
    :param source_files: locations of the files to be transferred
    :param destination_files: locations those files will be transferred to
    :param source_endpoint: id of the source endpoint
    :param destination_endpoint: id of the destination endpoint
    """
    # Validate source and destination paths
    if len(source_files) != len(destination_files):
        raise ValueError(
            f"Number of source and destination files must be equal."
            f"{len(source_files)} source files and {len(destination_files)} destination files were provided."
        )

    # Validate endpoints
    uuid.UUID(source_endpoint)
    uuid.UUID(destination_endpoint)

    tc = get_globus_transfer_client()
    transfer = create_transfer(tc, source_endpoint, destination_endpoint)

    # Add files to transfer
    for i, file in enumerate(source_files):
        transfer.add_item(source_files[i], destination_files[i])

    # Submit transfer
    transfer_result = tc.submit_transfer(transfer)

    while not tc.task_wait(transfer_result["task_id"], polling_interval=60):
        for event in tc.task_event_list(transfer_result["task_id"]):
            if event["is_error"]:
                raise GlobusError("Transfer unsuccessful")
